# 进行人脸录入 / Face register

import dlib
import numpy as np
import cv2
import os  # os库是py内置的标准库，可用于对文件和目录的的基本操作
import shutil
import time

from PIL import Image, ImageDraw, ImageFont

# 调用Dlib库的正向人脸检测器
detector = dlib.get_frontal_face_detector()
font_path = "./font/MSYH.TTC"
font_chinese = ImageFont.truetype(font_path, 30)


# 类名要使用驼峰命名
class FaceRegister:
    # __init__初始化，唯一方法，其中的self就类似Java中的this自身调用
    def __init__(self):
        self.path_photos_from_camera = "data/data_face/"
        self.font = cv2.FONT_ITALIC

        self.existing_faces_cnt = 0  # 已录入的人脸计数器 / cnt for counting saved faces
        self.ss_cnt = 0  # 录入 personX 人脸时图片计数器 / cnt for screen shots
        self.current_frame_faces_cnt = 0  # 录入人脸计数器 / cnt for counting faces in current frame

        self.save_flag = 1  # 之后用来控制是否保存图像的 flag / The flag to control if save
        self.press_n_flag = 0  # 之后用来检查是否先按 'n' 再按 's' / The flag to check if press 'n' before 's'

        # 初始化帧率、FPS
        self.frame_time = 0
        self.frame_start_time = 0
        self.fps = 0

    # 新建保存人脸图像文件和数据CSV文件夹
    def pre_work_mkdir(self):
        # 如果存在则跳过
        if os.path.isdir(self.path_photos_from_camera):
            pass
        else:
            os.mkdir(self.path_photos_from_camera)

    # 删除之前存的人脸数据文件夹
    def pre_work_del_old_face_folders(self):
        # 删除之前存的人脸数据文件夹, 删除 "/data_faces/person_x/" 中的png和csv文件
        folders_rd = os.listdir(self.path_photos_from_camera)
        for i in range(len(folders_rd)):
            shutil.rmtree(self.path_photos_from_camera + folders_rd[i])
        if os.path.isfile("data/features_all.csv"):
            os.remove("data/features_all.csv")

    # 如果有之前录入的人脸, 在之前 person_x 的序号按照 person_x+1 开始录入
    def check_existing_faces_cnt(self):
        if os.listdir("data/data_face/"):
            # 获取已录入的最后一个人脸序号
            person_list = os.listdir("data/data_face/")
            person_num_list = []
            for person in person_list:
                person_num_list.append(int(person.split('_')[-1]))
            self.existing_faces_cnt = max(person_num_list)

        # 如果第一次存储或者没有之前录入的人脸, 按照 person_1 开始录入
        else:
            self.existing_faces_cnt = 0

    # 获取处理之后 视频流 的帧数
    def update_fps(self):
        now = time.time()
        self.frame_time = now - self.frame_start_time
        self.fps = 3.0 / self.frame_time
        self.frame_start_time = now

    # 生成的 cv2 window 上面添加说明文字
    def draw_note(self, img_rd):
        # 显示在图片上
        img_pil = Image.fromarray(img_rd)
        draw = ImageDraw.Draw(img_pil)
        face_text = "脸部注册"
        # FPS_text = "帧率：" + str(self.fps.__round__(2))
        # draw.text((20, 40), face_text, font=font_chinese, fill=(255, 255, 255))
        # draw.text((20, 100), FPS_text, font=font_chinese, fill=(255, 255, 255))
        img_rd = np.array(img_pil)
        cv2.putText(img_rd, "Face Register", (20, 40), self.font, 1, (255, 255, 255), 1, cv2.LINE_AA)
        # cv2.putText(img_rd, "FPS: " + str(self.fps.__round__(2)), (20, 100), self.font, 0.8, (0, 255, 0), 1,
        #             cv2.LINE_AA)
        cv2.putText(img_rd, "Faces: " + str(self.current_frame_faces_cnt), (20, 140), self.font, 0.8, (0, 255, 0), 1,
                    cv2.LINE_AA)
        cv2.putText(img_rd, "N: Create face folder", (20, 350), self.font, 0.8, (255, 255, 255), 1,
                    cv2.LINE_AA)
        cv2.putText(img_rd, "S: Save current face", (20, 400), self.font, 0.8, (255, 255, 255), 1, cv2.LINE_AA)
        cv2.putText(img_rd, "Q: Quit", (20, 450), self.font, 0.8, (255, 255, 255), 1, cv2.LINE_AA)

    # 主要的 关键代码 获取人脸
    def process(self, stream):
        # 1. 新建储存人脸图像文件目录
        global current_face_dir
        self.pre_work_mkdir()
        # 2. 删除 "/data/data_faces_from_camera" 中已有人脸图像文件
        if os.path.isdir(self.path_photos_from_camera):
            self.pre_work_del_old_face_folders()

        # 3. 检查 "/data/data_faces_from_camera" 中已有人脸文件
        self.check_existing_faces_cnt()

        while stream.isOpened():
            flag, img_rd = stream.read()  # 获取摄像头的视屏流
            stdin = cv2.waitKey(1)  # 获取键盘键入
            faces = detector(img_rd, 0)  # 使用Dlib库的脸部探测器,0是不放大图片

            # 4. 按下 'n' 新建存储人脸的文件夹
            if stdin == ord('n'):
                self.existing_faces_cnt += 1
                current_face_dir = self.path_photos_from_camera + "person_" + str(self.existing_faces_cnt)
                os.makedirs(current_face_dir)
                print('\n')
                print("新建的人脸文件夹：", current_face_dir)

                self.ss_cnt = 0  # 将人脸计数器清零
                self.press_n_flag = 1  # 已经按下 'n'

            # 5. 检测到人脸
            if len(faces) != 0:
                # 矩形框
                for k, d in enumerate(faces):
                    # 计算矩形框大小
                    height = (d.bottom() - d.top())
                    width = (d.right() - d.left())
                    hh = int(height / 2)
                    ww = int(width / 2)

                    # 6. 判断人脸矩形框是否超出 480x640
                    if (d.right() + ww) > 640 or (d.bottom() + hh > 480) or (d.left() - ww < 0) or (d.top() - hh < 0):
                        cv2.putText(img_rd, "OUT OF RANGE", (20, 300), self.font, 0.8, (0, 0, 255), 1, cv2.LINE_AA)
                        color_rectangle = (0, 0, 255)
                        save_flag = 0
                        if stdin == ord('s'):
                            print("请调整位置！")
                    else:
                        color_rectangle = (155, 200, 11)  # 边框颜色
                        save_flag = 1
                    # 绘制边框
                    cv2.rectangle(img_rd,
                                  tuple([d.left() - ww, d.top() - hh]),
                                  tuple([d.right() + ww, d.bottom() + hh]),
                                  color_rectangle, 2)

                    # 7. 根据人脸大小生成空的图像 申请内存空间
                    img_blank = np.zeros((int(height * 2), width * 2, 3), np.uint8)

                    if save_flag:
                        # 8. 按下 's' 保存摄像头中的人脸到本地
                        if stdin == ord('s'):
                            # 检查有没有先按'n'新建文件夹 / Check if you have pressed 'n'
                            if self.press_n_flag:
                                self.ss_cnt += 1
                                for ii in range(height * 2):
                                    for jj in range(width * 2):
                                        img_blank[ii][jj] = img_rd[d.top() - hh + ii][d.left() - ww + jj]
                                cv2.imwrite(current_face_dir + "/img_face_" + str(self.ss_cnt) + ".jpg", img_blank)
                                print("写入本地图片信息：",
                                      str(current_face_dir) + "/img_face_" + str(self.ss_cnt) + ".jpg")
                            else:
                                print("请先按 'N' 新建文件夹,而后按'S'")

            self.current_frame_faces_cnt = len(faces)

            # 9. 在生成的窗口添加说明文字
            self.draw_note(img_rd)

            # 10. 按下 'q' 键退出 / Press 'q' to exit
            if stdin == ord('q'):
                break

            # 11. 控制帧率的刷新，可以关，关闭之后就不能在窗口看见 帧率
            self.update_fps()

            cv2.namedWindow("camera", 1)
            cv2.imshow("camera", img_rd)

    def run(self):
        cap = cv2.VideoCapture(0)  # 打开笔记本摄像头
        self.process(cap)  # 运行主方法

        cap.release()  # 释放资源
        cv2.destroyAllWindows()  # 关闭所有窗口


def main():
    face_register_con = FaceRegister()
    face_register_con.run()


if __name__ == '__main__':
    main()
